/**
 * GenerationMethods.cpp - Bitcoin address generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
/**
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
*/
 

#include "GenerationMethods.h"

#include "GenerationParameters.h"
#include "commandLine.h"
#include "util.h"
#include "random.h"
#include "privKey.h"
#include "bitcoinKeyPair.h"
#include "mnemonic.h"
#include "SaveFolderHandler.h"
#include "txtGeneration.h"
#include "qrenc.h"
#include "randomUtils.h"
#include "FileUtil.h"
#include "currencySpecific.h"
#include "keyDerivationFunction.h"
#include "bitgenVersion.h"
#include "printPS.h"
#include "ExtendedKey.h"
#include "hmac.h"
#include "bitgenConfig.h"
#include "pubXmrAddress.h"
#include "Directory.h"
#include "CliArguments.h"
#include "InformationList.h"
#include "SingleFolderHandler.h"


#include <iostream>
#include <list>
#include <stdlib.h>
#include <algorithm> 
#include <iterator>
#include <map>
#include <sstream>



void GenerationMethods::listSingleFolder(const CliArguments& args, const std::string& thetime)
{
	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen list FOLDER" << std::endl;
		std::cout << "FOLDER: pub, priv" << std::endl;
		
		throw std::string("Incorrect usage");
	}

	const std::string itemStr = args[0];
	const int listItem = GenerationParameters::parsePubPriv(itemStr);

	int paramStartPos = 1;
	bool listAll(false);

	if(args.size() > 1)
	{
		const std::string allStr = args[1];
		if(allStr != "all")
		{
			std::string errStr("Unexpected argument:_");
			errStr += allStr;
			throw errStr;
		}
		
		paramStartPos++;
		listAll = true;
	}
	
	if(listAll)
	{
		std::cout << "Will list all existing addresses, including empty amounts" << std::endl;
	}
	

	GenerationParameters params;
	CliArguments argsEnd = args.startAt(paramStartPos);
	params.parseArgs(argsEnd);
	params.checkNameProvided();
	
	std::string publicFolder;
	std::string privFolder;
	BitgenConfig::getThePublicAndPrivFolders(publicFolder, privFolder);


	if(listItem == -1)
	{
		if(!directoryExist(publicFolder))
		{
			std::cout << "PUblic directory does not exist: " << publicFolder << std::endl;
			throw std::string("Folder does not exist");
		}

		InformationList::listPublicFolderContents(publicFolder, thetime, params.getAccountNameInternal(), listAll);			
	}
	else if(listItem == -2)
	{		
		if(!directoryExist(privFolder))
		{
			std::cout << "Private directory does not exist: " << privFolder << std::endl;
			throw std::string("Folder does not exist");
		}

		InformationList::listPrivateFolderContents(privFolder, thetime, params.getAccountNameInternal(), listAll);			
	}	
}


void setSingleAsEmptyAmount(const std::string& pubFolder, const int addressNum, const std::string& thetime)
{
	const std::string pubEmptyFile = SingleFolderHandler::getSinglePubEmptyFile(pubFolder, addressNum);

	std::cout << "Writing file: " << pubEmptyFile << std::endl;

	const std::string contents = createEmptyFileContent(addressNum, thetime);

	fileWrite(pubEmptyFile, contents);
}



void GenerationMethods::markAsEmptyAmount(const CliArguments& args, const std::string& thetime)
{	
	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen empty NUMADDRESS" << std::endl;
		
		throw std::string("Incorrect usage");
	}

	const std::string numAddressStr = args[0];
	const int addressNum = atoi(numAddressStr.c_str());
	
	std::cout << "Setting as spent with amount zero for address: " << addressNum << std::endl;

	std::string publicFolder;
	std::string privFolder;
	BitgenConfig::getThePublicAndPrivFolders(publicFolder, privFolder);

	setSingleAsEmptyAmount(publicFolder, addressNum, thetime);
	setPrivFolderAsEmptyAmount(privFolder, addressNum, thetime);	
}




void GenerationMethods::generatePublicPart(const CliArguments& args, const std::string& thetime)
{
	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen pubaddr bitcoinAddress [name LABEL] [extraArgs]" << std::endl;
		return;			
	}
	
	const std::string adr(args[0]);
	const PubAddress myPub(adr);

	const CliArguments extraArgs = args.startAt(1);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);

	GenerationParameters params;
	
	params.parseArgs(extraArgs);
	params.checkNameProvided();
	
	if(params.getAccountNameInternal() != "")
	{
		std::cout << "Label: " << params.getAccountNameInternal() << std::endl;
	}
		
	const int newNum = GenerationMethods::getNextNumberIfConfigured(params);

	
	std::list<ExtraInfo> generationExtra;
	
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, "theGenMethod");

	const AddressGenerationInfo addrInfo(myPub.toString(), "", "", info);


	const Language lang = getLanguage();
	const PostscriptGenerator postscriptGenerator(info, versionNumber, applicationName, lang);
	const PublicPostscriptInfo pubInfo(addrInfo);
				
	const std::string psPublic = postscriptGenerator.postscriptPagePublic(pubInfo);
	
	const std::string pubFilename = "pub_" + (info.getNewNum() == 0 ? "" : (intToString(info.getNewNum()) + "_")) + myPub.toString().substr(0, 8) + "_share.ps";
	std::cout << "Saving publfile: " << pubFilename << std::endl;
	fileWrite(pubFilename, psPublic);	
}




void GenerationMethods::splitPassphrase(const std::string& passphrase, std::list<std::string>& generationExtra)
{	
	int lastNewlinePos(0);
	for(int i = 0 ; i < passphrase.size() ; i++)
	{
		const char c = passphrase[i];
		
		const int rowLength = i - lastNewlinePos;
				
		if(rowLength > 40 && c == ' ')
		{
			const std::string row = passphrase.substr(lastNewlinePos, i - lastNewlinePos);
			generationExtra.push_back(row);
			
			lastNewlinePos = i + 1;
		}
		else if(rowLength > 50)
		{
			const std::string row = passphrase.substr(lastNewlinePos, i - lastNewlinePos);
			generationExtra.push_back(row);
			
			lastNewlinePos = i;
		}
	}
	
	const std::string row = passphrase.substr(lastNewlinePos);
	generationExtra.push_back(row);
	
}








void GenerationMethods::printRandomSourceWarning(const std::string generator)
{
	std::cout << "Generating root key information" << std::endl << std::endl;
		
	std::cout << "Warning: Random number generated outside of bitgen with ";
	
#ifdef __linux	
	std::cout << generator << ".";
#else
	std::cout << "CryptGenRandom().";
#endif
	
	std::cout << std::endl;
	std::cout << "         Ensure that your random number generator gives good random numbers." << std::endl;
	std::cout << "         Flaws in the random number generation " << std::endl;
	std::cout << "         might make the private key possible to guess." << std::endl;
	std::cout << "         Consider other random number generators, such as dice." << std::endl;				
}





void GenerationMethods::generateBinaryRandomNumber(const CliArguments& args, const std::string& thetime)
{
	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen binary METHOD [extra1] [extra2]" << std::endl;
		std::cout << "    Method: random, dice, hex, hash, brain" << std::endl;
		
		throw std::string("Not enough arguments");
	}
	
	const std::string method = args[0];

	const std::string extra = ((args.size() > 1) ? args[1] : "");
	const std::string extra2 = ((args.size() > 2) ? args[2] : "");
	
	std::cout << "Generating 256 bit random number" << std::endl;
	
	const BigInt<256> privkey = generateRandomWithMethod(method, extra, extra2);

	std::cout << std::endl;
	std::cout << "Generated number: " << hexString(privkey.getBinaryLong()) << std::endl;
}








void GenerationMethods::saveTheFilesInFolders(const AddressGenerationInfo& allInfo, const std::string& privPath, const std::string& pubPath)
{
	
	if(allInfo.params.getGenerateTXT())
	{
		saveTxtFiles(pubPath, privPath, allInfo);		
	}
	
	if(allInfo.params.getGeneratePS())
	{		
		const Language lang = getLanguage();		
		const PostscriptGenerator postscriptGenerator(allInfo, versionNumber, applicationName, lang);
		
		const PublicPostscriptInfo pubInfo(allInfo);		
		const PrivatePostscriptInfo privInfo(allInfo);
				
		const PsInfoFile psinfo = postscriptGenerator.generate(privInfo, pubInfo, true);
			
		const std::string padExtra = (allInfo.getIsPad() ? "_pad" : "");
		const std::string addressStr = allInfo.pubaddr;

		const std::string numStr = (allInfo.getNewNum() == 0) ? "" : (intToString(allInfo.getNewNum()) + "_");


		const std::string filenameStart = getCurrencyShortName() + "_address_" + numStr + addressStr.substr(0, 8) + padExtra;

		
		const std::string privFilename = filenameStart + "_private.ps";
		const std::string fullPrivPath = (privPath == ".") ? privFilename : (privPath + "/" + privFilename);
	
		if(!allInfo.getIsPad())
		{
			std::cout << "Saving privfile: " << fullPrivPath << std::endl;
			fileWrite(fullPrivPath, psinfo.privatePs);
		}

		const std::string pubFilename = filenameStart + "_share.ps";
		const std::string fullPubPath = (pubPath == ".") ? pubFilename : (pubPath + "/" + pubFilename);	
		std::cout << "Saving publfile: " << fullPubPath << std::endl;
		fileWrite(fullPubPath, psinfo.publicPs);
		
	}
}




void GenerationMethods::saveAllFiles(const AddressGenerationInfo& allInfo, const bool useConf)
{	
	std::string privPath(".");
	std::string pubPath(".");

	const bool hasSingleFolderConfig = BitgenConfig::hasSingleFolderConfig();

	//Use the folder configuration if it exist
	if(hasSingleFolderConfig || useConf)
	{
		std::string confPublicFolder;
		std::string confPrivFolder;
		BitgenConfig::getThePublicAndPrivFolders(confPublicFolder, confPrivFolder);		
							
		SaveFolderHandler::createSingleNewFolders(allInfo.getNewNum(), confPublicFolder, confPrivFolder, pubPath, privPath);						
	}
	else if(allInfo.params.getUseFolder())	//If there is no folder configuration but the "folder" command line option is used
	{	
		SaveFolderHandler::createSingleNewFolders(allInfo.getNewNum(), ".", ".", pubPath, privPath);								
	}
	
	saveTheFilesInFolders(allInfo, privPath, pubPath);	
}




int GenerationMethods::getNextNumberIfConfigured(const GenerationParameters& params)
{
	if(params.getUseFolder())
	{
		const int newFolderNum = BitgenConfig::getConfiguredNum();
		//params.setAccountNameForNumber(params.getUseFolder(), newFolderNum);
		return newFolderNum;
	}
	else if(params.useConter())
	{
		return SaveFolderHandler::getNextNumber();
	}
	
	return 0;
}


int GenerationMethods::getNextSeedNumberIfConfigured(const GenerationParameters& params)
{
	if(params.useConter())
	{		
		return SaveFolderHandler::getNextSeedNumber();
	}
	
	return 0;
}



void GenerationMethods::generateWithDice(const CliArguments& args, const std::string& thetime)
{
	GenerationParameters params;
			
	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen (single) dice digits(1-6) [extraargs]" << std::endl;
		std::cout << " or" << std::endl;
		std::cout << "       bitgen dice - [AccountName]" << std::endl;
			
		return;
	}

	std::string dice(args[0]);


	const CliArguments argsEnd = args.startAt(1);	
	
	params.parseArgs(argsEnd);	
	params.checkNameProvided();
	
		
	if(dice == "-")
	{
		std::cout << "Enter the 100 dice values at one row as digits 1-6" << std::endl;
		std::cout << "Press enter when all digits are entered" << std::endl;
		
		dice = readStdinLine(100, true, oneToSix);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;			
		std::cout << std::endl;
	}
		
	if(dice.size() != 100)
	{
		std::string errMsg("Need exactly 100 dice throws, got: ");
		errMsg += intToString(dice.size());
		errMsg += " throws";
			
		throw errMsg;
		
		//std::cerr << "Need exactly 100 dice throws, got: " << dice.size() << " throws" << std::endl;
		//return;
	}

	
	const int newFolderNum = getNextNumberIfConfigured(params);


	std::cout << "Generating bitcoin address from dice, got " << dice.size() << " throws" << std::endl;
	std::cout << "Dice result (1-6): " << dice << std::endl;
		
	const BigInt<256> privkey = getRandomNumFrom6SideDice(dice);		
		
	std::list<ExtraInfo> generationExtra;
	const ExtraInfo extraInfo("Dice val (1-6)", dice);
	generationExtra.push_back(extraInfo);

	const PrivKey myPrivKey(privkey, params.getCompressed());
		
	const BitcoinKeyPair keypair = BitcoinKeyPair::calcKeys(myPrivKey);

	std::cout << "PUB: " << keypair.pubaddr << std::endl;
	std::cout << "PRIV: " << keypair.privaddr << std::endl;

	const Mnemonic mnemonicObj(privkey);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);	
	const GenerationInfoCore infoCore(infoBase, newFolderNum, params);
	const GenerationInfo info(infoCore, generationExtra, "dice");
	const AddressGenerationInfo allInfo(keypair.pubaddr.toString(), keypair.privaddr.toString(), "Wif", info);

	saveAllFiles(allInfo, params.useConf());
}





std::list<std::string> splitHashStr(const int totNum, const std::string& totalHashVal)
{
	//std::cout << "Will get hash num" << std::endl;
	const int totSize = totalHashVal.size();
	const int partSize = totSize / totNum;	//Rounded down
	const int useSize = partSize * totNum;

	std::list<std::string> ret;
	
	for(int i = 0 ; i < totNum ; i++)
	{
		//size must be a multiple of num, remove last characters if needed
		const std::string useHash = totalHashVal.substr(0, useSize);
	
		const int startPos = i * partSize;
	
		const std::string hashStr = useHash.substr(startPos, partSize);
		ret.push_back(hashStr); 
	}
	
	return ret;
}





int getLastNum(const std::map<int, EntryInfo>& info)
{
	int last(0);
		
	for(const std::pair<int, EntryInfo>& val : info)
	{
		last = val.first;
	}

	return last;
}









void GenerationMethods::generateRandom(const CliArguments& allExtraArgs, const std::string& thetime)
{
#ifndef __linux
	std::cerr << "Random is not supported on Windows. Use \"hash\" instead" << std::endl;
	return;
#endif

	ExtraOptionsRandom extraOptions;	

	GenerationParameters params;
	params.parseArgs(allExtraArgs, extraOptions);		
	params.checkNameProvided();
	
	
	if(params.getWords() != 24)
	{
		std::cout << "Num words: " << params.getWords() << std::endl;
		throw std::string("words argument not possible to use for address generation");
	}
	
	
	if(extraOptions.useURandom)
	{
		std::cout << "Using /dev/urandom with hash of buffer" << std::endl;		
	}


	const int newNum = getNextNumberIfConfigured(params);
	
	const std::string generator = extraOptions.useURandom ? "/dev/urandom" : "/dev/random";

	printRandomSourceWarning(generator);
		
	std::cout << "Generating random address" << std::endl;
	if(!extraOptions.useURandom)
	{
		std::cout << "Press random keys or move the mouse if needed" << std::endl;
	}
	
	const BigInt<256> privKey = extraOptions.useURandom ? getBlockCalcRandomNumber() : getRandomNumber(extraOptions.longbuf);
		
	const PrivKey myPrivKey(privKey, params.getCompressed());

		
	const std::string hexStr = hexString(privKey.getBinaryLong());

	std::list<ExtraInfo> generationExtra;
	
	const ExtraInfo extraInfo("Raw hex", hexStr);
	generationExtra.push_back(extraInfo);
	
	BitcoinEllipticCurve ec;

	const Wif wif = myPrivKey.getWif();
	const PubKey pubKey = myPrivKey.getPubkey(ec, extraOptions.onlyIndicatedUsage, extraOptions.signVerify);
	const PubAddress pubAddr = pubKey.getAddress();
	
	std::cout << "PUB : " << pubAddr << std::endl;			
	std::cout << "PRIV: " << wif.toString() << std::endl;
	std::cout << std::endl;

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);				
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, generator);
	
	const AddressGenerationInfo allInfo(pubAddr.toString(), wif.toString(), "Wif", info);
			
	saveAllFiles(allInfo, params.useConf());
}






void GenerationMethods::generateFromWif(const CliArguments& args, const std::string& thetime)
{
	if(args.size() < 1)
	{
		std::cout << "Error, no wif given" << std::endl;
		std::cout << "Usage: bitgen wif WIF" << std::endl;
		return;
	}

	const std::string wifStr(args[0]);

	const CliArguments argsEnd = args.startAt(1);
	
	GenerationParameters params;
	params.parseArgs(argsEnd);
	params.checkNameProvided();
	
	const int newNum = getNextNumberIfConfigured(params);
		


	const Wif wif(wifStr);
	
	const PrivKey myPrivKey = wif.getPriv();

	std::list<ExtraInfo> generationExtra;
	
	const ExtraInfo extraInfo("Wif", wif.toString());
	generationExtra.push_back(extraInfo);
	
	BitcoinEllipticCurve ec;

	const PubKey pubKey = myPrivKey.getPubkey(ec);
	const PubAddress pubAddr = pubKey.getAddress();
	
	std::cout << "PUB : " << pubAddr << std::endl;
	std::cout << "PRIV: " << wif.toString() << std::endl;
	std::cout << std::endl;
				
	const Mnemonic mnemonicObj(myPrivKey.key);		
		
	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);						
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, "Wif");
	const AddressGenerationInfo allInfo(pubAddr.toString(), wif.toString(), "Wif", info);


	saveAllFiles(allInfo, params.useConf());	
}




std::string GenerationMethods::generateOneFileContents(const BitcoinKeyPair& keypair, const bool addPubkey, std::string& customString)
{
	std::string contents;
	
	if(!customString.empty())
	{
		for(int i = 0 ; i < customString.size() ; i++)
		{
			const char c = customString[i];
			if(c == '#')		//Pub
			{
				contents += keypair.pubaddr.toString();				
			}
			else if(c == '$')		//CashAddr Pub
			{
				contents += keypair.pubaddr.toCashAddrString();				
			}			
			else if(c == '@')	//Priv
			{
				contents += keypair.privaddr.toString();
			}
			else
			{
				contents += c;
			}
		}

		contents += "\n";

		return contents;
	}
	
	contents += keypair.pubaddr.toString();
	contents += ", ";
	if(addPubkey)
	{
		contents += keypair.pubaddr.toString();
		contents += ", ";							
	}
	contents += keypair.privaddr.toString();
	contents += "\n";
	
	return contents;
}






GenerationParameters parseBulkParameters(const CliArguments& args, bool& useURandom, bool& longbuf, bool& oneFile, bool& addPubkey, bool& readCustom, 
	int& num, std::string& customString)
{
	GenerationParameters params;

	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen bulk NumAddresses [extraOptions]" << std::endl;
		throw std::string("Incorrect usage");
	}

	const std::string numAddresses(args[0]);
	//std::cout << "NUM ADDR: " << numAddresses << std::endl;
	num = atoi(numAddresses.c_str());

	const CliArguments allExtraArgs = args.startAt(1);
	
	for(int i = 0 ; i < allExtraArgs.size() ; i++)	
	{
		const std::string& arg = allExtraArgs[i];
		
		if(readCustom)
		{
			readCustom = false;
			customString = arg;
			std::cout << "Custom string: " << customString << std::endl;
			continue;
		}
		
		if(params.useArg(arg))
		{
			//Nothing
		}						
		else if(arg == "urandom")
		{
			if(longbuf)
			{
				throw std::string("Can not combine urandom and long");
			}
			useURandom = true;
		}
		else if(arg == "long")
		{
			if(useURandom)
			{
				throw std::string("Can not combine urandom and long");
			}
			
			std::cout << "Using 128 bytes from /dev/random" << std::endl;
			longbuf = true;
		}
		else if(arg == "onefile")
		{
			oneFile = true;
		}
		else if(arg == "onefilecustom")
		{
			oneFile = true;
			readCustom = true;
		}		
		else if(arg == "pubkey")
		{
			addPubkey = true;
		}		
		else
		{
			std::cout << "Incorrect argument: " << arg << std::endl;
			throw std::string("Argument error");			
		}
	}
	
	params.checkNameProvided();
	
	return params;
}




void GenerationMethods::generateBulk(const CliArguments& args, const std::string& thetime)
{		
	bool useURandom(false);
	bool longbuf(false);	
	bool oneFile(false);
	bool addPubkey(false);
	bool readCustom(false);

	int num(-1);

	std::string customString;

	const GenerationParameters params = parseBulkParameters(args, useURandom, longbuf, oneFile, addPubkey, readCustom, num, customString);

	const std::string generator = useURandom ? "/dev/urandom" : "/dev/random";


	std::cout << "Generating " << num << " random addresses" << std::endl;
	
	if(!useURandom)
	{
		std::cout << "Press random keys or move the mouse if needed" << std::endl;
	}

	std::string contents;
	std::string filename;

	for(int i = 0 ; i < num ; i++)
	{
		std::cout << "Address " << (i + 1) << " " << std::flush;

		const int newNum = getNextNumberIfConfigured(params);

		
		const BigInt<256> privKey = useURandom ? getBlockCalcRandomNumber() : getRandomNumber(longbuf);
				
		const PrivKey myPrivKey(privKey, params.getCompressed());
		const BitcoinKeyPair keypair = BitcoinKeyPair::calcKeys(myPrivKey);

		std::cout << keypair.pubaddr.toString() << std::endl;
		
		if(oneFile)
		{
			filename = std::string("bulk_") + keypair.pubaddr.toString() + ".txt";
			
			contents += generateOneFileContents(keypair, addPubkey, customString);			
		}
		else
		{	
			
			const std::string hexStr = hexString(privKey.getBinaryLong());

			std::list<ExtraInfo> generationExtra;
			const ExtraInfo extraInfo("Raw hex", hexStr);
			generationExtra.push_back(extraInfo);
						
			const Mnemonic mnemonicObj(privKey);

			const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);						
			const GenerationInfoCore infoCore(infoBase, newNum, params);
			const GenerationInfo info(infoCore, generationExtra, generator);
			const AddressGenerationInfo allInfo(keypair.pubaddr.toString(), keypair.privaddr.toString(), "Wif", info);
			
			saveAllFiles(allInfo);
		}
	}

	if(oneFile)
	{
		std::cout << std::endl;
		std::cout << "Writing file: " << filename << std::endl;
		fileWrite(filename, contents);

	}
		
	std::cout << std::endl;
	std::cout << "All addresses generated. Ready" << std::endl;
}










void GenerationMethods::generateHash(const CliArguments& args, const std::string& thetime)
{
	GenerationParameters params;
	const int minChars(200);
	
	std::cout << "Generating address from hash" << std::endl;

	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen (single) hash \"The hashval\" [extraOptions]" << std::endl;
		std::cout << " or" << std::endl;
		std::cout << "       bitgen hash - [extraOptions]" << std::endl;
		
		return;			
	}

	std::string hashval(args[0]);

	const CliArguments argsEnd = args.startAt(1);

	params.parseArgs(argsEnd);
	params.checkNameProvided();

	
	if(hashval == "-")
	{
		std::cout << "Enter the hash string at one row" << std::endl;
		std::cout << "Press enter when everything is entered" << std::endl;
		
		hashval = readStdinLine(minChars, false, everythingOk);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;			
		std::cout << std::endl;		
	}


	const int newNum = getNextNumberIfConfigured(params);

	const BigInt<256> privKey = hashStringToBigInt(hashval, minChars);
		
	std::list<ExtraInfo> generationExtra;
	const std::string privHex = hexString(privKey.getBinaryLong());

	const ExtraInfo extraInfo("Hex from hash", privHex);
	generationExtra.push_back(extraInfo);
			
	const PrivKey myPrivKey(privKey, params.getCompressed());

	const BitcoinKeyPair keypair = BitcoinKeyPair::calcKeys(myPrivKey);

	const Mnemonic mnemonicObj(privKey);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, "hash");

	const AddressGenerationInfo allInfo(keypair.pubaddr.toString(), keypair.privaddr.toString(), "Wif", info);

	saveAllFiles(allInfo, params.useConf());
}









void GenerationMethods::generateHex(const CliArguments& args, const std::string& thetime)
{
	GenerationParameters params;
		
	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen-" << getCurrencyShortName() << " hex hexdigits [extraOptions]" << std::endl;
		std::cout << " or" << std::endl;
		std::cout << "       bitgen-" << getCurrencyShortName() << " hex - [extraOptions]" << std::endl;
		
		return;			
	}

	std::string privKeyStrMix(args[0]);

		
	const CliArguments argsEnd = args.startAt(1);
	
	params.parseArgs(argsEnd);
	params.checkNameProvided();
		

	if(privKeyStrMix == "-")
	{
		std::cout << "Enter the 64 hex values at one row" << std::endl;
		std::cout << "Press enter when all digits are entered" << std::endl;
		
		privKeyStrMix = readStdinLine(64, true, hexChar);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;
		std::cout << std::endl;
	}
	
	
	const BigInt<256> privKey = hexStringToBigInt(privKeyStrMix);

	const int newNum = getNextNumberIfConfigured(params);


	std::cout << "Generating bitcoin address from hex private key" << std::endl;

	std::cout << "Private key (hex): " << hexString(privKey.getBinaryLong()) << std::endl;	

	std::list<ExtraInfo> generationExtra;
	const ExtraInfo extraInfo("Hex privkey", hexString(privKey.getBinaryLong()));
	generationExtra.push_back(extraInfo);				
	
	BitcoinEllipticCurve ec;

	const PrivKey myPrivKey(privKey, params.getCompressed());
	const PubKey pubKey = myPrivKey.getPubkey(ec);
	const PubAddress address = pubKey.getAddress();

	if(params.printInfo())
	{
		std::cout << "Wif    : " << myPrivKey.toString() << std::endl;
		std::cout << "Pkif   : " << pubKey.toString() << std::endl;		
		std::cout << "Address: " << address.toString() << std::endl;
		std::cout << "CashAddress: " << address.toCashAddrString() << std::endl;		
	}

	const Mnemonic mnemonicObj(privKey);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);		
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, "hex input");
	const AddressGenerationInfo allInfo(address.toString(), myPrivKey.toString(), "Wif", info);
		
	saveAllFiles(allInfo, params.useConf());
}








bool GenerationMethods::dispatchSingle(const CliArguments& args, const std::string& thetime, const bool requireMethodSelect)
{
	if(requireMethodSelect)
	{
	if(args.size() < 1)
	{
		std::cout << "Usage: bitgen " << args.prefixString() << "gentype" << std::endl;
		throw std::string("Incorrect usage");		
		return false;
	}
	}

	const std::string genType(args[0]);
	//std::cout << "Gentype: " << genType << std::endl;

	const CliArguments argsEnd = args.startAt(1);

	
	if(genType == "random")
	{
		std::cout << "Using random generation" << std::endl;
		GenerationMethods::generateRandom(argsEnd, thetime);
		return true;		
	}
	else if(genType == "brainwallet")
	{
		std::cout << "Using brainwallet generation" << std::endl;
		GenerationMethods::generateBrainWallet(argsEnd, thetime);
		return true;
	}
	else if(genType == "dice")
	{
		std::cout << "Using dice generation" << std::endl;
		GenerationMethods::generateWithDice(argsEnd, thetime);
		return true;
	}
	else if(genType == "hash")
	{
		std::cout << "Using hash generation" << std::endl;		
		GenerationMethods::generateHash(argsEnd, thetime);
		return true;
	}
	else if(genType == "hex")
	{
		std::cout << "Using hash generation" << std::endl;
		GenerationMethods::generateHex(argsEnd, thetime);
		return true;
	}
	else if(genType == "wif")
	{
		std::cout << "Using wif generation" << std::endl;
		GenerationMethods::generateFromWif(argsEnd, thetime);		
		return true;
	}
	else if(genType == "bulk")
	{
		std::cout << "Using bulk generation" << std::endl;
		GenerationMethods::generateBulk(argsEnd, thetime);		
		return true;
	}	
	else
	{
		if(requireMethodSelect)
		{
			std::cout << "Unknown gentype: " << genType << std::endl;
			throw std::string("Unknown gentype");
		}
		
		return false;
	}	
	
	return false;
}




//#ifndef NO_BRAINWALLET_GENERATION

void GenerationMethods::generateBrainWallet(const CliArguments& args, const std::string& thetime)
{
#ifdef NO_BRAINWALLET_GENERATION
		std::cout << "No brainwallet support" << std::endl;
		return;
#else		
	
		
		
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen (single) brainwallet email/salt \"The passphrase\" [extraOptions]" << std::endl;
		return;
	}

	const std::string salt(args[0]);
	std::cout << "Salt: " << salt << std::endl;

	std::string passphrase(args[1]);
	std::cout << "Passphrase: " << passphrase << std::endl;


	const CliArguments argsEnd = args.startAt(2);
		
	GenerationParameters params;
	params.parseArgs(argsEnd);
	params.checkNameProvided();
	


	if(passphrase == "-")
	{
		std::cout << "Enter the brainwallet phrase" << std::endl;
		std::cout << "Press enter when all characters are entered" << std::endl;
		
		passphrase = readStdinLine(50, false, everythingOk);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;			
		std::cout << std::endl;		
	}

	const int newNum = getNextNumberIfConfigured(params);

	std::cout << "Generating brainwallet" << std::endl;

	std::cout << "Salt: \"" << salt << "\"" << std::endl;		
	std::cout << "Passphrase: \"" << passphrase << "\"" << std::endl;


	const BigInt<256> privKey =  brainwalletToBigInt(passphrase, salt);
		
	std::list<ExtraInfo> generationExtra;
	const ExtraInfo extraInfo("Passphrase", passphrase);
	generationExtra.push_back(extraInfo);		
		
		
	std::cout << "Calculating keys..." << std::endl;	
		
	const PrivKey myPrivKey(privKey, params.getCompressed());
		
	const BitcoinKeyPair keypair = BitcoinKeyPair::calcKeys(myPrivKey);

	const Mnemonic myMnemonicObj(privKey);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, "brainwallet");
	const AddressGenerationInfo allInfo(keypair.pubaddr.toString(), keypair.privaddr.toString(), "Wif", info);


	saveAllFiles(allInfo, params.useConf());

#endif
		
}









BigInt<256> generateRandomWithMethod(const std::string& method, const std::string& extra, const std::string& extra2)
{
	if(method == "random")
	{
		std::cout << "Generating from /dev/random" << std::endl;

		const std::string random = getFromDevRandom(32);
		std::cout << std::endl;
	
		const BigInt<256> privkey(random, 'c');
	
	
		//std::cout << "Number: " << hexString(privkey.getBinaryLong()) << std::endl;
		return privkey;
	}
	else if(method == "dice")
	{
		std::cout << "Generating from dice: " << extra << std::endl;
		
		const BigInt<256> privkey = getRandomNumFrom6SideDice(extra);		
	
		return privkey;
	}
	else if(method == "hex")
	{
		std::cout << "Generating from hex: " << extra << std::endl;
		
		const BigInt<256> privkey = hexStringToBigInt(extra);
		
		return privkey;
	}
	else if(method == "hash")
	{
		std::cout << "Generating from hash: " << extra << std::endl;
				
		const BigInt<256> privkey = hashStringToBigInt(extra);
		
		return privkey;
	}
	
	else if(method == "brain")
	{
#ifdef NO_BRAINWALLET_GENERATION
		std::cout << "No brainwallet support" << std::endl;
		throw std::string("Brainwallet not supported");
#else				
		std::cout << "Brain string: " << extra << std::endl;
		std::cout << "Salt: " << extra2 << std::endl;
		std::cout << std::endl;
		
		const BigInt<256> privkey =  brainwalletToBigInt(extra, extra2);
						
		return privkey;
#endif
	}	
	else
	{
		std::cout << "Unknown method: " << method << std::endl;
		throw std::string("Unknown method");
	}
}


